/*
	Sudoku generator

	Batch sudoku generator

	(C) Copyright 2005 by Mogens Kjaer, mk@lemo.dk

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <limits.h>
#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#include "sudoku.h"

int nfilled=28;
int bestndigits=0;
int ndigits;
int symmetrical=1;
int level=LEVEL_EASY;
int thislevel=0;
int maxndigits=1000000;
int maxlevel=999;
int minlevel=0;
int nsudoku;
int generatedsudokus=0;
/*                 1234567890123456789012345678901234567890123456789012345678901234567890 */
char symbollist[]="0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

int statistics[MAXNXNY*MAXNXNY][NLEVELS];
static time_t start_time, stop_time;

int origdisplayboard[MAXNXNY][MAXNXNY];
int displayboard[MAXNXNY][MAXNXNY];
int solutionboard[MAXNXNY][MAXNXNY];

int setboard[MAXNXNY][MAXNXNY];
int setboardsave[MAXNXNY][MAXNXNY];
int beforeboard[MAXNXNY][MAXNXNY];
int board[MAXNXNY][MAXNXNY];
int initialpossible[MAXNXNY][MAXNXNY];
int possible[MAXNXNY][MAXNXNY];
int eliminated[MAXNXNY][MAXNXNY];
int nx=3;
int ny=3;
int nxny=9;
int nsol=0;
int ntest=0;
int stop_pressed=0;
int letter=0;
int print_explain=1;
int progstat=0;
int solution_unique=0;

int nlines=0;
char *loglines=NULL;
int logsize=0;

void loginit()
{
  if(loglines!=NULL) free(loglines);
  loglines=NULL;
  logsize=0;
  nlines=0;
}

void logend()
{
}

void logprintf(const char *template, ...)
{
  va_list ap;
  char tmpbuf[1024];
  int len,i;

  va_start(ap, template);
  vsprintf(tmpbuf, template, ap);
  va_end(ap);
  len=strlen(tmpbuf);
  for(i=0;i<len;i++)
  {
    if(tmpbuf[i]=='\n') nlines++;
  }

  if(loglines==NULL)
  {
    loglines=malloc(len+1);
    loglines[0] = '\0';
  }
  else
  {
    loglines=realloc(loglines, strlen(loglines)+len+1);
  }
  strcat(loglines, tmpbuf);
}

void update_menu(char *txt)
{
}

int test_stop()
{
  return 0;
}

int a2toi(char *c)
{
  return((c[0]-'0')*10+(c[1]-'0'));
}

int genabslevel(char *explanation)
{
  int abslevel,elen,i,n;
  char type;

  abslevel=0;

  elen=strlen(explanation);

  for(i=0; i<elen;)
  {
    type=explanation[i];
    if(type=='A')
    {
      abslevel += 1;
      i += 6;
    }
    else if(type=='B')
    {
      abslevel += nxny;
      i += 8;
    }
    else if(type=='C')
    {
      abslevel += nxny;
      i += 8;
    }
    else if(type=='D')
    {
      abslevel += nxny;
      i += 10;
    }
    else if(type=='E')
    {
      abslevel += nxny*ny;
      i += 8;
    }
    else if(type=='F')
    {
      abslevel += nxny*nx;
      i += 8;
    }
    else if(type=='G')
    {
      abslevel += nxny*ny;
      i += 10;
    }
    else if(type=='H')
    {
      abslevel += nxny*nx;
      i += 10;
    }
    else if(type=='I')
    {
      n = a2toi(explanation+i+1);
      abslevel += nxny*nxny*n*n;
      i += n*5+3;
    }
    else if(type=='J')
    {
      i += 6;
    }
    else if(type=='K')
    {
      i += 6;
    }
    else if(type=='L')
    {
      i += 6;
    }
    else if(type=='M')
    {
      n = a2toi(explanation+i+1);
      abslevel += nxny*nxny*nxny*n;
      i += n+7;
    }
    else if(type=='N')
    {
      i += 6;
    }
    else if(type=='O')
    {
      n = a2toi(explanation+i+1);
      i += n+7;
    }
    else if(type=='Y')
    {
      i += 1;
    }
    else if(type=='Z')
    {
      i += 1;
    }
    else
    {
      fprintf(stderr, "Unknown explanation code: %c\n", type);
    }
  }
  return abslevel;
}

void main_update()
{
  int i,j,abslevel;
  if(bestndigits>maxndigits || thislevel>maxlevel || thislevel<minlevel) return;
  generatedsudokus++;
  printf("insert into sudoku values(NULL,now(),%d,%d,%d,%d,%d,0,0);\n",nx,ny,bestndigits,thislevel,symmetrical);
  printf("insert into problems values(LAST_INSERT_ID(),'");
  for(i=0;i<nxny;i++)
  for(j=0;j<nxny;j++) printf("%c", symbollist[displayboard[i][j]]);
  printf("');\n");
  printf("insert into solutions values(LAST_INSERT_ID(),'");
  loginit();
  solve_sudoku(1);
  logend();
  for(i=0;i<nxny;i++)
  for(j=0;j<nxny;j++) printf("%c", symbollist[abs(displayboard[i][j])]);
  printf("');\n");
  printf("insert into explanations values(LAST_INSERT_ID(),'%s');\n", loglines);
  abslevel=genabslevel(loglines);
  printf("update sudoku set abslevel=%d where serial=LAST_INSERT_ID();\n", abslevel);
  fflush(stdout);
}

main(int argc, char **argv)
{
  int isudoku;
  char *env;
  unsigned int seed;
  int serial,i,j;
  char *problem;

  env=getenv("SUDOKU_SEED");
  if(env)
  {
    seed=atoi(env);
  }
  else
  {
    seed=time(NULL);
  }

  srandom(seed);

  if(argc>1)
  {
    nsudoku=atoi(argv[1]);
    if(nsudoku>0)
    {
      if(argc>2)
      {
	nx=atoi(argv[2]);
	if(argc>3)
	{
	  ny=atoi(argv[3]);
	  if(argc>4)
	  {
	    symmetrical=atoi(argv[4]);
            if(argc>5)
	    {
	      maxndigits=atoi(argv[5]);
	      if(argc>6)
	      {
	        maxlevel=atoi(argv[6]);
		if(argc>7)
		{
		  minlevel=atoi(argv[7]);
		}
		else
		{
		  minlevel=0;
		}
	      }
	      else
	      {
	        maxlevel=999;
	      }
	    }
	    else
	    {
	      maxndigits=100000;
	    }
	  }
	  else
	  {
	    symmetrical=1;
	  }
	}
	else
	{
	  ny=3;
	}
      }
      else
      {
	nx=3;
      }
    }
    else
    {
      if(argc<5)
      {
        fprintf(stderr, "Usage: bsudoku nsudoku nx ny symmetry maxdigits maxlevel minlevel\nor: bsudoku 0 nx ny serial problem\n");
	exit(-1);
      }
      nx=atoi(argv[2]);
      ny=atoi(argv[3]);
      serial=atoi(argv[4]);
      problem=argv[5];
    }
  }
  else
  {
    nsudoku=100;
  }
  if(nx<2) nx=2;
  if(ny<2) ny=2;
  if(nx>MAXNX)
  {
    fprintf(stderr, "nx is too big. The maximum is: %d\n", MAXNX);
    exit(-1);
  }
  if(ny>MAXNY)
  {
    fprintf(stderr, "ny is too big. The maximum is: %d\n", MAXNY);
    exit(-1);
  }
  if(symmetrical<0) symmetrical=0;
  if(symmetrical>=NSYMMETRICALS) symmetrical=NSYMMETRICALS-1;
  nxny=nx*ny;

  init_bitmaps(nxny);

  nfilled=1;
  level = LEVEL_ANY;

  if(nsudoku>0)
  {
    progstat = 1;
    for(isudoku=0;isudoku<nsudoku;isudoku++)
    {
      bestndigits=0;
      sudoku_run();
    }
    fprintf(stderr, "Generated a total of %d sudokus.\n", generatedsudokus);
  }
  else
  {
    /* Solve a given sudoku and print out explanation */
    for(i=0; i<nxny; i++)
    for(j=0; j<nxny; j++)
    {
      if(*problem == '\0')
      {
        fprintf(stderr, "Problem does not match nx and ny\n");
	exit(-1);
      }
      if(*problem >= '0' && *problem <= '9')
      {
        displayboard[i][j] = *problem - '0';
      }
      else if(*problem >= 'A' && *problem <= 'Z')
      {
        displayboard[i][j] = *problem - 'A' + 10;
      }
      else
      {
        fprintf(stderr, "Invalid character %c in problem\n", *problem);
	exit(-1);
      }
      problem++;
    }
    if(*problem != '\0')
    {
      fprintf(stderr, "Problem does not match nx and ny\n");
      exit(-1);
    }
    if(nsudoku<0)
    {
      /* Refine sudoku; try to insert a digit to see
         if we can remove others */
      for(i=0; i<nxny; i++)
      for(j=0; j<nxny; j++)
      board[i][j]=displayboard[i][j];
      sudoku_refine();
    }
    else
    {
      loginit();
      if(solve_sudoku(1)!=1)
      {
	fprintf(stderr, "Problem does not have a unique solution.\n");
	exit(-1);
      }
      logend();

      printf("update sudoku set explanation='%s' where serial=%d;\n", loglines, serial);
    }

  }
}
