/*
	Sudoku generator

	Motif interface

	(C) Copyright 2005 by Mogens Kjaer, mk@lemo.dk

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <time.h>

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/MainW.h>
#include <Xm/RowColumn.h>
#include <Xm/MessageB.h>
#include <Xm/FileSB.h>
#include <Xm/DialogS.h>
#include <Xm/Separator.h>
#include <Xm/ToggleB.h>
#include <Xm/Protocols.h>
#include <Xm/DrawingA.h>
#include <Xm/ScrolledW.h>
#include <Xm/TextF.h>
#include <Xm/CutPaste.h>
#include <Xm/Separator.h>

#include "sudoku.h"

static Widget main_wnd=NULL, printer_wnd=NULL, settings_wnd=NULL;
static Widget toplevel;
static Widget main_form, settings_pb, copy_pb, paste_pb, create_pb, solve_pb, play_pb, hint_pb, print_pb, quit_pb,about_pb;
static Widget drawing_wnd=NULL;
static Widget main_settingsdialog, settings_form, 
	      nx_label, nx_radio, nxradios[MAXNX+1],
	      ny_label, ny_radio, nyradios[MAXNY+1],
	      settings_label, settings_edit,
	      separator1, separator2, separator3,
              settings_radio2, symmetricals[NSYMMETRICALS],
	      settings_radio, levels[NLEVELS],
	      check1, check2,
	      settings_smallinitial, settings_autosmall,
              settings_ok, settings_cancel;
static Widget main_printdialog, printer_fileform, printer_fileselection, papersize_label, papersize,
              papersize_pane, A4_w, Letter_w, printer_explain_label, printer_explain_check,
	      printer_explain_select;

typedef struct
{
  XFontStruct *textfont;
  XFontStruct *smalltextfont;
  XFontStruct *smallboldtextfont;
  Pixel redcolor, blackcolor, graycolor, pinkcolor;
} ApplicationData, *ApplicationDataPtr;

static ApplicationData applicationdata;

static XtResource resources[] =
{
  {"textfont", "Textfont", XtRFontStruct, sizeof(XFontStruct *),
   XtOffset(ApplicationDataPtr, textfont),
   XtRString, "-ibm-courier-medium-r-normal--40-*-*-*-m-*-iso8859-1"
  },
  {"smalltextfont", "SmallTextfont", XtRFontStruct, sizeof(XFontStruct *),
   XtOffset(ApplicationDataPtr, smalltextfont),
   XtRString, "-ibm-courier-medium-r-normal--18-*-*-*-m-*-iso8859-1"
  },
  {"smallboldtextfont", "SmallBoldTextfont", XtRFontStruct, sizeof(XFontStruct *),
   XtOffset(ApplicationDataPtr, smallboldtextfont),
   XtRString, "-ibm-courier-bold-r-normal--18-*-*-*-m-*-iso8859-1"
  },
  {"redcolor", "Redcolor", XtRPixel, sizeof(Pixel),
   XtOffset(ApplicationDataPtr, redcolor),
   XtRString, "red"
  },
  {"blackcolor", "Blackcolor", XtRPixel, sizeof(Pixel),
   XtOffset(ApplicationDataPtr, blackcolor),
   XtRString, "black"
  },
  {"graycolor", "Graycolor", XtRPixel, sizeof(Pixel),
   XtOffset(ApplicationDataPtr, graycolor),
   XtRString, "gray"
  },
  {"pinkcolor", "Pinkcolor", XtRPixel, sizeof(Pixel),
   XtOffset(ApplicationDataPtr, pinkcolor),
   XtRString, "pink"
  },
};

static GC thick_gc, thin_gc, black_gc, red_gc, background_gc;
static GC small_black_gc, small_gray_gc, small_pink_gc;
static int font_width,font_height;
static int smallfont_width,smallfont_height;

int nfilled=28;
int bestndigits=0;
int ndigits;
int symmetrical=1;
int level=LEVEL_EASY;
int thislevel=0;
int smallinitial=1;
int autosmall=1;
static int new_smallinitial, new_autosmall;
static int newlevel=LEVEL_EASY;
static int newsymmetrical=1;
static int newnx=3;
static int newny=3;
static char filled_buffer[4];
char symbollist[]="0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

int statistics[MAXNXNY*MAXNXNY][NLEVELS];
static time_t start_time, stop_time;
static time_t play_start_time, play_stop_time;
static int stat_nfilled,stat_nempty,stat_nbad,playtime;

#define THINWIDTH 1
#define THICKWIDTH 3
#define THICKOFFSET 1
#define SUBCELLSIZE 18

int displayboard[MAXNXNY][MAXNXNY];
int origdisplayboard[MAXNXNY][MAXNXNY];
int solutionboard[MAXNXNY][MAXNXNY];

int setboard[MAXNXNY][MAXNXNY];
int setboardsave[MAXNXNY][MAXNXNY];
int beforeboard[MAXNXNY][MAXNXNY];
int board[MAXNXNY][MAXNXNY];
int initialpossible[MAXNXNY][MAXNXNY];
int possible[MAXNXNY][MAXNXNY];
int eliminated[MAXNXNY][MAXNXNY];
int nx=3;
int ny=3;
int nxny=9;
int nsol=0;
int ntest=0;
int stop_pressed=0;
int letter=0;
int print_explain=1;
/*
	Program status: progstat:

	0:	Just started
	1:	Finding puzzles
	2:	Puzzle found
	3:	Playing
	4:	Showing solution
	5:	Showing solution after play is finished
*/
int progstat=0;
int solution_unique=0;

int n;
Arg wargs[1000];

int nlines=0;
char *loglines=NULL;
int logsize=0;

void loginit()
{
  if(loglines!=NULL) free(loglines);
  loglines=NULL;
  logsize=0;
  nlines=0;
}

void logend()
{
}

void logprintf(const char *template, ...)
{
  va_list ap;
  char tmpbuf[1024];
  int len,i;

  va_start(ap, template);
  vsprintf(tmpbuf, template, ap);
  va_end(ap);
  len=strlen(tmpbuf);
  nlines++;
  
  if(loglines==NULL)
  {
    loglines=malloc(len+1);
    loglines[0] = '\0';
  }
  else
  {
    loglines=realloc(loglines, strlen(loglines)+len+1);
  }
  strcat(loglines, tmpbuf);
}

void setpbtext(Widget w, char *txt)
{
  XmString stmp;

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateSimple(txt))); n++;
  XtSetValues(w, wargs, n);

  XmStringFree(stmp);
}

void setpbsensitivity(Widget w, int sensitive)
{
  n=0;
  XtSetArg(wargs[n], XmNsensitive, sensitive?True:False); n++;
  XtSetValues(w, wargs, n);
}

void settitle(char *txt)
{
  XTextProperty text_prop;

  text_prop.value = (unsigned char *) txt;
  text_prop.encoding = XA_STRING;
  text_prop.format = 8;
  text_prop.nitems = strlen(txt);
  XSetWMName(XtDisplay(toplevel), XtWindow(toplevel), &text_prop);
}

void update_menu()
{
  char windowtitle[200];
  switch(progstat)
  {
  case 0:
    setpbsensitivity(copy_pb, 0);
    setpbsensitivity(paste_pb, 1);
    setpbtext(create_pb, "Create");
    setpbsensitivity(solve_pb, 0);
    setpbtext(solve_pb, "Show solution");
    setpbsensitivity(play_pb, 0);
    setpbtext(play_pb, "Play");
    setpbsensitivity(hint_pb, 0);
    setpbsensitivity(print_pb, 0);
    sprintf(windowtitle, "Sudoku generator.");
    break;
  case 1:
    setpbsensitivity(copy_pb, 0);
    setpbsensitivity(paste_pb, 0);
    setpbtext(create_pb, "Stop!");
    setpbsensitivity(solve_pb, 0);
    setpbtext(solve_pb, "Show solution");
    setpbsensitivity(play_pb, 0);
    setpbtext(play_pb, "Play");
    setpbsensitivity(hint_pb, 0);
    setpbsensitivity(print_pb, 0);
    if(bestndigits==0)
    {
      sprintf(windowtitle, "Sudoku generator. No puzzle found so far.");
    }
    else
    {
      sprintf(windowtitle, "Sudoku generator. %d cells in %d iteration%s", bestndigits, ntest, ntest==1?"":"s");
    }
    break;
  case 2:
    setpbsensitivity(copy_pb, 1);
    setpbsensitivity(paste_pb, 1);
    setpbtext(create_pb, "Create");
    setpbsensitivity(solve_pb, 1);
    setpbtext(solve_pb, "Show solution");
    setpbsensitivity(play_pb, 1);
    setpbtext(play_pb, "Play");
    setpbsensitivity(hint_pb, 0);
    setpbsensitivity(print_pb, 1);
    sprintf(windowtitle, "Sudoku generator. Found puzzle with %d cells in %d iteration%s.", bestndigits, ntest, ntest==1?"":"s");
    break;
  case 3:
    setpbsensitivity(copy_pb, 1);
    setpbsensitivity(paste_pb, 1);
    setpbtext(create_pb, "Create");
    setpbsensitivity(solve_pb, 1);
    setpbtext(solve_pb, "Show solution");
    setpbsensitivity(play_pb, 1);
    setpbtext(play_pb, "Stop play");
    setpbsensitivity(hint_pb, 1);
    setpbsensitivity(print_pb, 1);
    if(stat_nbad==0)
    {
      sprintf(windowtitle, "Sudoku Generator. %d empty cell%s", stat_nempty, stat_nempty==1?"":"s");
    }
    else
    {
      sprintf(windowtitle, "Sudoku Generator. %d empty cell%s, %d cell%s wrong", stat_nempty, stat_nempty==1?"":"s",
	    stat_nbad, stat_nbad>1?"s are":" is");
    }
    break;
  case 4:
    setpbsensitivity(copy_pb, 1);
    setpbsensitivity(paste_pb, 1);
    setpbtext(create_pb, "Create");
    setpbsensitivity(solve_pb, 1);
    setpbtext(solve_pb, "Hide solution");
    setpbsensitivity(play_pb, 1);
    setpbtext(play_pb, "Play");
    setpbsensitivity(hint_pb, 0);
    setpbsensitivity(print_pb, 1);
    sprintf(windowtitle, "Sudoku generator. Solution.");
    break;
  case 5:
    setpbsensitivity(copy_pb, 1);
    setpbsensitivity(paste_pb, 1);
    setpbtext(create_pb, "Create");
    setpbsensitivity(solve_pb, 1);
    setpbtext(solve_pb, "Hide solution");
    setpbsensitivity(play_pb, 1);
    setpbtext(play_pb, "Play");
    setpbsensitivity(hint_pb, 0);
    setpbsensitivity(print_pb, 1);
    sprintf(windowtitle, "Sudoku generator. Solved in %d:%2.2d:%2.2d.", playtime/3600, (playtime/60)%60, playtime%60);
    break;
  }
  settitle(windowtitle);
}

static void printer_print()
{
}

static void main_clear()
{
  int i,j;
  for(i=0;i<nxny;i++)
  for(j=0;j<nxny;j++)
  displayboard[i][j] = 0;
}

void main_update()
{
  char txt[16];
  int row,col,xsize,ysize,x,y,xline,width,thickoffset;
  int row2,col2,value;

  xsize = nxny*SUBCELLSIZE*ny+(nx-1)*THICKWIDTH+(nxny-nx)*THINWIDTH;
  ysize = nxny*SUBCELLSIZE*nx+(ny-1)*THICKWIDTH+(nxny-ny)*THINWIDTH;

  y=0;
  XDrawLine(XtDisplay(drawing_wnd), XtWindow(drawing_wnd), thin_gc, 0, 0, xsize, 0);
    
  y=THINWIDTH;
  xline=0;
  for(row=0;row<nxny;row++)
  {
    if(row>0)
    {
      thickoffset = (row%ny==0)?THICKOFFSET:0;
      XDrawLine(XtDisplay(drawing_wnd), XtWindow(drawing_wnd), (row%ny==0)?thick_gc:thin_gc,
		xline+thickoffset, 0, xline+thickoffset, ysize);
      thickoffset = (row%nx==0)?1:0;
      XDrawLine(XtDisplay(drawing_wnd), XtWindow(drawing_wnd), (row%nx==0)?thick_gc:thin_gc,
                0, y+thickoffset, xsize, y+thickoffset);
      y+=(row%nx==0)?THICKWIDTH:THINWIDTH;
      xline+=(row%ny==0)?THICKWIDTH:THINWIDTH;
    }
    x=0;
    for(col=0;col<nxny;col++)
    {
      if(col>0) x+=(col%ny==0)?THICKWIDTH:THINWIDTH;
      txt[1]='\0';
      XFillRectangle(XtDisplay(drawing_wnd), XtWindow(drawing_wnd), background_gc, x, y, SUBCELLSIZE*ny, SUBCELLSIZE*nx);
      if(progstat!=3 || displayboard[row][col]>0)
      {
	if(displayboard[row][col])
	{
	  txt[0]=symbollist[abs(displayboard[row][col])];
	}
	else
	{
	  txt[0]=' ';
	}
	XDrawImageString(XtDisplay(drawing_wnd), XtWindow(drawing_wnd),
	    displayboard[row][col]<0?red_gc:black_gc,
	    x+SUBCELLSIZE*ny/2-font_width/2,
	    y+SUBCELLSIZE*nx-(SUBCELLSIZE*nx-applicationdata.textfont->ascent)/2,
	    txt, strlen(txt));
      }
      else
      {
        for(row2=0; row2<nx; row2++)
	for(col2=0; col2<ny; col2++)
	{
	  value=row2*ny+col2+1;
	  if(initialpossible[row][col] & (1<<value))
	  {
	    txt[0]=symbollist[value];
	    XDrawImageString(XtDisplay(drawing_wnd), XtWindow(drawing_wnd),
	        (eliminated[row][col] & (1<<value))?small_pink_gc:
		(possible[row][col] & (1<<value))?small_black_gc:small_gray_gc,
		x+col2*SUBCELLSIZE+SUBCELLSIZE/2-smallfont_width/2,
		y+row2*SUBCELLSIZE+SUBCELLSIZE-(SUBCELLSIZE-applicationdata.smalltextfont->ascent)/2,
		txt, strlen(txt));
	  }
	}
      }
      x+=SUBCELLSIZE*ny;
    }
    y+=SUBCELLSIZE*nx;
    xline+=SUBCELLSIZE*ny;
  }
  XDrawLine(XtDisplay(drawing_wnd), XtWindow(drawing_wnd), thin_gc,
	    xline, 0, xline, ysize);
  XDrawLine(XtDisplay(drawing_wnd), XtWindow(drawing_wnd), thin_gc,
	    0, y, xsize, y);
}

static void generate_sudoku()
{
  bestndigits=0;
  ntest=0;
  stop_pressed=0;
  main_clear();
  main_update();
  progstat=1;
  update_menu();
}

static void settings_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  int i;
  if(progstat != 1)
  {
    newlevel=level;
    newsymmetrical=symmetrical;
    newnx=nx;
    newny=ny;
    new_smallinitial=smallinitial;
    new_autosmall=autosmall;
    sprintf(filled_buffer, "%d", nfilled);
    n=0;
    XtSetArg(wargs[n], XmNvalue, filled_buffer); n++;
    XtSetValues(settings_edit, wargs, n);
    for(i=0; i<NLEVELS; i++)
    {
      n=0;
      XtSetArg(wargs[n], XmNset, (newlevel==i)?True:False); n++;
      XtSetValues(levels[i], wargs, n);
    }
    for(i=0; i<NSYMMETRICALS; i++)
    {
      n=0;
      XtSetArg(wargs[n], XmNset, (newsymmetrical==i)?True:False); n++;
      XtSetValues(symmetricals[i], wargs, n);
    }
    for(i=2; i<MAXNX; i++)
    {
      n=0;
      XtSetArg(wargs[n], XmNset, (newnx==i)?True:False); n++;
      XtSetValues(nxradios[i], wargs, n);
    }
    for(i=2; i<MAXNY; i++)
    {
      n=0;
      XtSetArg(wargs[n], XmNset, (newny==i)?True:False); n++;
      XtSetValues(nyradios[i], wargs, n);
    }

    n=0;
    XtSetArg(wargs[n], XmNset, new_smallinitial?True:False); n++;
    XtSetValues(settings_smallinitial, wargs, n);

    n=0;
    XtSetArg(wargs[n], XmNset, new_autosmall?True:False); n++;
    XtSetValues(settings_autosmall, wargs, n);

    XtRealizeWidget(main_settingsdialog); 
  }
}

static void copy_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  char textboard[2048];
  char symbol[2];
  int i,r,c;
  long item_id = 0;
  int status;
  XmString clip_label;
  static int cnt=0;
  Display *dpy = XtDisplayOfObject(w);
  Window window = XtWindowOfObject(w);

  cnt++;

  clip_label = XmStringCreateLocalized("Sudoku");

  textboard[0] = '\0';
  symbol[1] = '\0';

  for(r=0; r<nxny; r++)
  {
    if(r>0 && (r%nx)==0)
    {
      for(c=0; c<nxny; c++)
      {
	if(c>0 && (c%ny)==0) strcat(textboard, "-+");
        strcat(textboard, "--");
      }
      strcat(textboard, " \n");
    }
    for(c=0; c<nxny; c++)
    {
      if(c>0 && (c%ny)==0) strcat(textboard, " |");
      strcat(textboard, " ");
      if(displayboard[r][c]>0)
      {
	symbol[0] = symbollist[displayboard[r][c]];
      }
      else
      {
        symbol[0] = '.';
      }
      strcat(textboard, symbol);
    }
    strcat(textboard, " \n");
  }

  do
    status = XmClipboardStartCopy(dpy, window, clip_label, CurrentTime, NULL, NULL, &item_id);
  while(status == ClipboardLocked);

  XmStringFree(clip_label);

  do
    status = XmClipboardCopy(dpy, window, item_id, "STRING",
             textboard, (long) strlen(textboard)+1, cnt, NULL);
  while(status == ClipboardLocked);

  do
    status = XmClipboardEndCopy(dpy, window, item_id);
  while(status==ClipboardLocked);
}

static void paste_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  int status;
  unsigned long received;
  char *data, buf[1024];
  Display *dpy = XtDisplayOfObject (w);
  Window window = XtWindowOfObject (w);
  int i,r,c,formatok;
  char ch;
  int newboard[MAXNXNY][MAXNXNY];
  if(progstat != 1)
  {
    if(progstat==4 || progstat==5)
    {
      unsolve_sudoku();
      main_update();
      progstat=2;
      update_menu();
    }
    do
      status = XmClipboardStartRetrieve (dpy, window, CurrentTime);
    while (status == ClipboardLocked);
    status = XmClipboardRetrieve (dpy, window, "STRING", buf, sizeof(buf)-1, &received, NULL);
    buf[received] = 0; /* NULL terminate */
    XmClipboardEndRetrieve (dpy, window);
    if (status == ClipboardSuccess)
    {
      formatok=1;
      i=0;
      r=0;
      c=0;
      while(ch=buf[i++])
      {
	if(ch=='.' || (ch>='0' && ch<='9') || (ch>='A' && ch<='Z'))
	{
	  if(ch=='.') ch='0';
	  if(r>=nxny)
	  {
	    formatok=0;
	  }
	  else
	  {
	    if(ch>='A')
	    {
	      newboard[r][c] = ch-'A'+10;
	    }
	    else
	    {
	      newboard[r][c] = ch-'0';
	    }
	    c++;
	    if(c==nxny)
	    {
	      c=0;
	      r++;
	    }
	  }
	}
      }
      if(r!=nxny || c!=0) formatok=0;
      bestndigits=0;
      for(r=0;r<nxny;r++)
      for(c=0;c<nxny;c++)
      {
	displayboard[r][c] = formatok?newboard[r][c]:0;
	if(displayboard[r][c]) bestndigits++;
      }
      ntest=1;
      progstat=formatok?2:0;
      main_update();
      update_menu();
    }
  }
}

static void update_play_status()
{
  int r1,c1;

  stat_nfilled=0;
  stat_nempty=0;
  stat_nbad=0;

  if(progstat==3)
  {
    for(r1=0;r1<nxny;r1++)
    for(c1=0;c1<nxny;c1++)
    {
      if(displayboard[r1][c1]==0)
      {
	stat_nempty++;
      }
      else
      {
	stat_nfilled++;
	if(displayboard[r1][c1]!=solutionboard[r1][c1]) stat_nbad++;
      }
    }
    if(stat_nempty==0 && stat_nbad==0)
    {
      progstat=5;
      for(r1=0;r1<nxny;r1++)
      for(c1=0;c1<nxny;c1++)
      {
        if(origdisplayboard[r1][c1]==0)
	{
	  beforeboard[r1][c1] = origdisplayboard[r1][c1];
	  displayboard[r1][c1] = -solutionboard[r1][c1];
	}
      }
      time(&play_stop_time);
      playtime = play_stop_time - play_start_time;
    }
  }
  update_menu();
}

static void create_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  int i,j,ok,total;
  double speed;
  FILE *statfile;
  if(progstat==1)
  {
    if(bestndigits>0)
    {
      progstat=2;
    }
    else
    {
      progstat=0;
    }
    stop_pressed=1;
    time(&stop_time);
    statfile = fopen("sudoku.stat", "w");
    if(statfile != NULL)
    {
      total=0;
      fprintf(statfile, "Filled\tVEasy\tEasy\tMedium\tHard\tDifficult\n");
      for(i=0; i<MAXNXNY*MAXNXNY; i++)
      {
	ok=0;
	for(j=0;j<NLEVELS;j++)
	{
	  if(statistics[i][j]>0) ok=1;
	}
	if(ok)
	{
	  fprintf(statfile, "%2d", i);
	  for(j=0;j<5;j++)
	  {
	    fprintf(statfile, "\t%d", statistics[i][j]);
	    total += statistics[i][j];
	  }
	  fprintf(statfile, "\n");
	}
      }
      speed=stop_time-start_time;
      fprintf(statfile, "%d Sudokus created in %d seconds.\n", total, (int) (stop_time-start_time));
      fprintf(statfile, "Speed: %.3f sudokus created per second.\n", ((double)total)/speed);
      fclose(statfile);
    }
  }
  else
  {
    for(i=0; i<MAXNXNY*MAXNXNY; i++)
    for(j=0; j<NLEVELS; j++) statistics[i][j]=0;
    time(&start_time);
    progstat=1;
    generate_sudoku();
  }
  update_menu();
}

static int OKPressed, Wnd_destroyed;

static void ok_callback(Widget w, XtPointer call_data, XtPointer client_data)
{
  OKPressed = TRUE;
} 

static void destroy_callback(Widget w, XtPointer call_data, XtPointer client_data)
{
  Wnd_destroyed = TRUE;
}

void informuser(Widget refwnd,char *txt)
{
  Widget inform_shell, form;
  Widget stext, ok_pb;
  int x, y;
  XmString stmp;

  OKPressed = FALSE;
  Wnd_destroyed = FALSE;

  /* Create a popup shell for the information window */
  n = 0;
  XtSetArg(wargs[n], XtNtitle, "Inform"); n++;
  XtSetArg(wargs[n], XmNmessageString, (stmp=XmStringCreateLocalized(txt))); n++;
  XtSetArg(wargs[n], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL); n++;
  XtSetArg(wargs[n], XmNautoUnmanage, FALSE); n++;
  XtSetArg(wargs[n], XmNnoResize, TRUE); n++;
  XtSetArg(wargs[n], XmNresizePolicy, XmRESIZE_NONE); n++;
  XtSetArg(wargs[n], XmNdeleteResponse, XmDESTROY); n++;
  inform_shell = XmCreateInformationDialog(refwnd, "inform", wargs, n);
  XmStringFree(stmp);

  XtUnmanageChild(XmMessageBoxGetChild(inform_shell, XmDIALOG_CANCEL_BUTTON));
  XtUnmanageChild(XmMessageBoxGetChild(inform_shell, XmDIALOG_HELP_BUTTON));
  				
  XtAddCallback(inform_shell, XmNdestroyCallback, destroy_callback, NULL);
  XtAddCallback(inform_shell, XmNokCallback, ok_callback, NULL);

  XtManageChild(inform_shell);

  /* Local event loop - wait for ok */
  while (!OKPressed)
  {
    XEvent event;

    XtNextEvent(&event);
    XtDispatchEvent(&event);
  }

  /* Popdown and destroy the information window */
  if (!Wnd_destroyed)
  {
    XtUnmanageChild(inform_shell);
    XtDestroyWidget(inform_shell);
  }
}

static void solve_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  char inform[120];
  if(progstat!=1)
  {
    if(progstat==4 || progstat==5)
    {
      unsolve_sudoku();
      progstat=2;
      main_update();
      update_menu();
    }
    else
    {
      stop_pressed=0;
      solution_unique=solve_sudoku(1);
      if(solution_unique==0)
      {
        sprintf(inform, "This sudoku does not have a solution.");
      }
      else if(solution_unique>1)
      {
        sprintf(inform, "This sudoku has more than one solution.");
      }
      else
      {
        sprintf(inform, "This sudoku has one solution only. Difficulty: ");
	switch(thislevel)
	{
        case LEVEL_VERYEASY:
	  strcat(inform, "Very easy");
	  break;
        case LEVEL_EASY:
	  strcat(inform, "Easy");
	  break;
        case LEVEL_MEDIUM:
	  strcat(inform, "Medium");
	  break;
        case LEVEL_HARD:
	  strcat(inform, "Hard");
	  break;
        case LEVEL_DIFFICULT:
	  strcat(inform, "Difficult");
	  break;
	}
      }
      informuser(toplevel, inform);
      progstat=4;
      main_update();
      update_menu();
    }
  }
}

static void hint_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  int olddisplayboard[MAXNXNY][MAXNXNY];
  int oldpossible[MAXNXNY][MAXNXNY];
  int i,j,badr,badc,nbad;
  char txt[200];
  char *buf1, *ploglines, *buf2, code;

  if(progstat==3)
  {
    nbad=0;
    for(i=0;i<nxny;i++)
    {
      for(j=0;j<nxny;j++)
      {
	if(displayboard[i][j]!=0 && displayboard[i][j]!=solutionboard[i][j]) nbad++;
	olddisplayboard[i][j] = displayboard[i][j];
	oldpossible[i][j] = possible[i][j];
      }
    }
    loginit();
    solve_sudoku(1);
    logend();
    for(i=0;i<nxny;i++)
    for(j=0;j<nxny;j++)
    {
      displayboard[i][j] = olddisplayboard[i][j];
      possible[i][j] = oldpossible[i][j];
    }
    progstat=3;
    if(loglines && (nbad==0))
    {
      if(loglines[0] == 'E' ||
         loglines[0] == 'F' ||
         loglines[0] == 'G' ||
         loglines[0] == 'H' ||
         loglines[0] == 'I' ||
         loglines[0] == 'J' ||
         loglines[0] == 'M')
      {
	buf2=NULL;
	ploglines=loglines;
	do
	{
	  code = ploglines[0];
	  i=logtext_en(ploglines, &buf1);
	  ploglines += i;
	  if(buf2==NULL)
	  {
	    buf2 = malloc(strlen(buf1)+1);
	    strcpy(buf2, buf1);
	  }
	  else
	  {
	    buf2 = realloc(buf2, strlen(buf1)+1+strlen(buf2)+1);
	    strcat(buf2, "\n");
	    strcat(buf2, buf1);
	  }
	} while(code != 'K' && code != 'L' && code != 'N');
      }
      else
      {
	logtext_en(loglines, &buf1);
	buf2=malloc(strlen(buf1)+1);
	strcpy(buf2, buf1);
      }
      informuser(toplevel, buf2);
      free(buf2);
    }
    else
    {
      /* No hints, maybe a mistake somewhere? */
      badr=-1;
      badc=-1;
      for(i=0;i<nxny;i++)
      for(j=0;j<nxny;j++)
      {
	if(displayboard[i][j] && displayboard[i][j]!=solutionboard[i][j])
	{
	  badr=i+1;
	  badc=j+1;
	  break;
	}
      }
      if(badr != -1)
      {
        sprintf(txt, "You've made a mistake in row %d, column %d.\n", badr, badc);
	informuser(toplevel, txt);
      }
      else
      {
	informuser(toplevel, "No hints");
      }
    }
  }
}

static void play_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  int i,j;
  if(progstat>1)
  {
    if(progstat==3)
    {
      progstat=2;
      for(i=0;i<nxny;i++)
      for(j=0;j<nxny;j++)
      displayboard[i][j] = origdisplayboard[i][j];
      update_play_status();
      main_update();
    }
    else
    {
      stop_pressed=0;
      if(progstat==4 || progstat==5)
      {
	unsolve_sudoku();
      }
      initplay(smallinitial);
      progstat=3;
      update_play_status();
      main_update();
      time(&play_start_time);
      stop_pressed=0;
    }
  }
}

static void print_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  if(progstat>=2)
  {
    XmFileSelectionDoSearch(printer_fileselection, NULL);
    XtRealizeWidget(main_printdialog);
  }
}

static void quit_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  exit(0);
}

static void about_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  char txt[1024];

  sprintf(txt, "Sudoku version %d, Copyright (C) 2006 by Mogens Kjr\n\nThis Sudoku program comes with ABSOLUTELY NO WARRANTY.\nThis is free software, and you are welcome to redistribute it under certain conditions.\n\nPlease see http://www.lemo.dk/sudoku", SUDOKU_VERSION);
  informuser(toplevel, txt);
}

static void xy2rc(int x, int y, int *r1, int *c1, int *r2, int *c2)
{
  int r,c,x1,x2,y1,y2;

  *r1 = -1;
  *c1 = -1;
  *r2 = -1;
  *c2 = -1;

  x1=0;
  for(c=0;c<nxny;c++)
  {
    x2=x1+SUBCELLSIZE*ny-1;
    if(x>=x1 && x<=x2)
    {
      *c1 = c;
      *c2 = (x-x1)/SUBCELLSIZE;
      break;
    }
    if(c>0) x1 += (c%ny==0)?THICKWIDTH:THINWIDTH;
    x1 += SUBCELLSIZE*ny;
  }

  y1=0;
  for(r=0;r<nxny;r++)
  {
    y2=y1+SUBCELLSIZE*nx-1;
    if(y>=y1 && y<=y2)
    {
      *r1 = r;
      *r2 = (y-y1)/SUBCELLSIZE;
      break;
    }
    if(r>0) y1 += (r%nx==0)?THICKWIDTH:THINWIDTH;
    y1 += SUBCELLSIZE*nx;
  }
}

static void drawing_input_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmDrawingAreaCallbackStruct *cb = (XmDrawingAreaCallbackStruct *) call_data;
  int r1,c1,r2,c2,value,update_needed;
  int i,j;

  update_needed = 0;

  if(progstat==3)
  {
    if(cb->event->type == ButtonRelease)
    {
      xy2rc(cb->event->xbutton.x, cb->event->xbutton.y, &r1, &c1, &r2, &c2);
      if(r1 != -1 && c1 != -1)
      {
        value = r2*ny+c2+1;
	if(cb->event->xbutton.button != 1)
	{
	  /* Right or middle button */
	  if(displayboard[r1][c1] == 0)
	  {
	    if(initialpossible[r1][c1] & (1<<value))
	    {
	      eliminated[r1][c1] ^= (1<<value);
	      update_needed = 1;
	    }
	  }
	}
	else
	{
	  /* Left button */
	  if(origdisplayboard[r1][c1]==0)
	  {
	    if(displayboard[r1][c1]==0)
	    {
	      if((possible[r1][c1] & (1<<value)) &&
		 !(eliminated[r1][c1] & (1<<value)))
	      {
	        displayboard[r1][c1] = value;
		update_needed = 1;
	      }
	    }
	    else
	    {
	      displayboard[r1][c1] = 0;
	      update_needed = 1;
	    }
	  }
	}
      }
    }
    if(update_needed)
    {
      if(autosmall)
      {
	for(i=0;i<nxny;i++)
	for(j=0;j<nxny;j++)
	{
	  board[i][j] = displayboard[i][j];
	}
	findpossible();
      }
      update_play_status();
      main_update();
    }
  }
  else
  {
    /* printf("Ignored event in non-playing mode.\n"); */
  }
}

static void drawing_expose_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  main_update();
}

static void level_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  newlevel = (long) client_data;
}

static void symmetrical_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  newsymmetrical = (long) client_data;
}

static void set_defaultnfilled()
{
  int newnxny,defaultnfilled,i;
  newnxny=newnx*newny;
  defaultnfilled=0;
  if(newnxny==4) defaultnfilled=8;
  else if(newnxny==6) defaultnfilled=12;
  else if(newnxny==8) defaultnfilled=20;
  else if(newnxny==9) defaultnfilled=28;
  else if(newnxny==10) defaultnfilled=36;
  else if(newnxny==12) defaultnfilled=56;
  else if(newnxny==15) defaultnfilled=100;
  else if(newnxny==16) defaultnfilled=112;
  else if(newnxny==20) defaultnfilled=200;
  else if(newnxny==25) defaultnfilled=300;
  if(defaultnfilled>0)
  {
    sprintf(filled_buffer, "%d", defaultnfilled);
    n=0;
    XtSetArg(wargs[n], XmNvalue, filled_buffer); n++;
    XtSetValues(settings_edit, wargs, n);
  }
  if(newnxny==4 && newlevel!=LEVEL_VERYEASY)
  {
    newlevel=LEVEL_VERYEASY;
    for(i=0; i<NLEVELS; i++)
    {
      n=0;
      XtSetArg(wargs[n], XmNset, (newlevel==i)?True:False); n++;
      XtSetValues(levels[i], wargs, n);
    }
  }
}

static void nx_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  newnx = (long) client_data;
  set_defaultnfilled();
}

static void ny_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  newny = (long) client_data;
  set_defaultnfilled();
}

static void recreate_drawing()
{
  if(drawing_wnd != NULL) XtDestroyWidget(drawing_wnd);
  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopWidget, settings_pb); n++;
  XtSetArg(wargs[n], XmNwidth, nxny*SUBCELLSIZE*ny+(nx-1)*THICKWIDTH+(nxny-nx-1)*THINWIDTH); n++;
  XtSetArg(wargs[n], XmNheight, nxny*SUBCELLSIZE*nx+(ny-1)*THICKWIDTH+(nxny-ny)*THINWIDTH); n++;
  drawing_wnd = XtCreateManagedWidget("drawing_wnd", xmDrawingAreaWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(drawing_wnd, XmNexposeCallback, drawing_expose_cb, (XtPointer) NULL);
  XtAddCallback(drawing_wnd, XmNinputCallback, drawing_input_cb, (XtPointer) NULL);
}

static void smallinitial_cb(Widget w, XtPointer client_data, XtPointer call_data)
{  
  XmToggleButtonCallbackStruct *tgb = (XmToggleButtonCallbackStruct *) call_data;
  int value,i; 

  value=True;
#if XmVERSION==1
  if(!tgb->set) value=False;
#else
  if(tgb->set == XmUNSET) value=False;
#endif
  new_smallinitial=value;
}

static void autosmall_cb(Widget w, XtPointer client_data, XtPointer call_data)
{  
  XmToggleButtonCallbackStruct *tgb = (XmToggleButtonCallbackStruct *) call_data;
  int value,i; 

  value=True;
#if XmVERSION==1
  if(!tgb->set) value=False;
#else
  if(tgb->set == XmUNSET) value=False;
#endif
  new_autosmall=value;
}

static void settingsok_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  char *s;

  s = XmTextFieldGetString(settings_edit);
  sscanf(s, "%d", &nfilled);
  XFree(s);

  symmetrical=newsymmetrical;
  level=newlevel;
  nx=newnx;
  ny=newny;
  nxny=nx*ny;
  autosmall=new_autosmall;
  smallinitial=new_smallinitial;

  init_bitmaps(nxny);
  
  /* Recreate main drawing area */
  recreate_drawing();

  main_clear();
  bestndigits=0;
  progstat=0;
  main_update();
  update_menu();

  XtUnrealizeWidget(main_settingsdialog);
}

static void settingscancel_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  XtUnrealizeWidget(main_settingsdialog);
}

static void printer_fileopen_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  XmFileSelectionBoxCallbackStruct *cbs = (XmFileSelectionBoxCallbackStruct *)call_data;
  char *filename;
  FILE *print_fh;
  double dev_xmin, dev_xmax, dev_ymin, dev_ymax, margin, yspacing;
  char *paper_text;
  double size, printer_cellsize, x, y, fontsize, width, printer_thinwidth, printer_thickwidth,
         smallfontsize;
  int row, col, i, j, npages, row2, col2, value, ilen;
  char *buf1;

  XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &filename);

  n=0;
  print_explain=0;
  stop_pressed=0;
  XtSetArg(wargs[n], XmNset, &print_explain); n++;
  XtGetValues(printer_explain_select, wargs, n);

  print_fh = fopen(filename, "w");
  if(print_fh != NULL)
  {
    if(letter)
    {
      dev_xmin = 0;
      dev_xmax = 612;
      dev_ymin = 0;
      dev_ymax = 792;
      margin = 18;
      paper_text = "Letter";
    }
    else
    {
      dev_xmin = 0;
      dev_xmax = 595;
      dev_ymin = 0;
      dev_ymax = 842;
      margin = 18;
      paper_text = "A4";
    }
    
    fprintf(print_fh, "%%!PS-Adobe-3.0\n");
    fprintf(print_fh, "%%%%Creator: Sudoku generator\n");
    fprintf(print_fh, "%%%%Pages: (atend)\n");
    fprintf(print_fh, "%%%%DocumentMedia: %s %.0f %.0f 0 white ()\n", paper_text, dev_xmax, dev_ymax);
    fprintf(print_fh, "%%%%BoundingBox: %.0f %.0f %.0f %.0f\n", dev_xmin, dev_ymin, dev_xmax, dev_ymax);
    fprintf(print_fh, "%%%%Orientation: Portrait\n");
    fprintf(print_fh, "%%%%EndComments\n\n");

    fprintf(print_fh, "[{\n");
    fprintf(print_fh, "%%%%BeginFeature: *PageSize %s\n", paper_text);
    fprintf(print_fh, "<</PageSize[%d %d]/ImagingBBox null>>setpagedevice\n", (int) dev_xmax, (int) dev_ymax);
    fprintf(print_fh, "%%%%EndFeature\n");
    fprintf(print_fh, "} stopped cleartomark\n");

    if((dev_xmax-dev_xmin)>(dev_ymax-dev_ymin))
    {
      size = dev_ymax-dev_ymin;
    }
    else
    {
      size = dev_xmax-dev_xmin;
    }

    size = size-2*margin;

    if((progstat==4 || progstat==5)&& print_explain && solution_unique==1)
    {
      size = size/3;
      fprintf(print_fh, "gsave %g %g translate\n", (dev_xmax-dev_xmin)/2-size/2, dev_ymax-margin-size);
    }
    else
    {
      fprintf(print_fh, "%g %g translate\n", (dev_xmax-dev_xmin)/2-size/2, (dev_ymax-dev_ymin)/2-size/2);
    }

    printer_cellsize=size/nxny;
    fontsize = printer_cellsize*0.9;
    printer_thinwidth = printer_cellsize/40;
    printer_thickwidth = printer_thinwidth*3;
    smallfontsize = fontsize/nx;
    if(ny>nx) smallfontsize=fontsize/ny;

    fprintf(print_fh, "/showcenter {dup gsave 0 0 moveto true charpath flattenpath pathbbox grestore /symax exch def /sxmax exch def /symin exch def /sxmin exch def sxmin sxmax sub 2 div sxmin sub symin symax sub 2 div rmoveto show} def\n");
    fprintf(print_fh, "/Courier findfont %g scalefont setfont\n", fontsize);

    y=size;
    for(row=0;row<nxny;row++)
    {
      if(row>0)
      {
	fprintf(print_fh, "0 0 0 setrgbcolor\n");
        width = (row%ny==0)?printer_thickwidth:printer_thinwidth;
	fprintf(print_fh, "%g setlinewidth\n", width);
	fprintf(print_fh, "newpath\n");
	fprintf(print_fh, "%g 0 moveto\n", y);
	fprintf(print_fh, "%g %g lineto\n", y, size);
	fprintf(print_fh, "stroke\n");
        width = (row%nx==0)?printer_thickwidth:printer_thinwidth;
	fprintf(print_fh, "newpath\n");
	fprintf(print_fh, "%g setlinewidth\n", width);
	fprintf(print_fh, "0 %g moveto\n", y);
	fprintf(print_fh, "%g %g lineto\n", size, y);
	fprintf(print_fh, "stroke\n");
      }
      x=0;
      for(col=0; col<nxny; col++)
      {
	if(progstat!=3 || displayboard[row][col]!=0)
	{
	  if(displayboard[row][col]!=0)
	  {
	    fprintf(print_fh, "/Courier findfont %g scalefont setfont\n", fontsize);
	    fprintf(print_fh, "%g %g moveto\n", x+printer_cellsize/2,y-printer_cellsize/2);
	    fprintf(print_fh, "%d 0 0 setrgbcolor\n", displayboard[row][col]<0?1:0);
	    fprintf(print_fh, "(%c) showcenter\n", symbollist[abs(displayboard[row][col])]);
	  }
	}
	else
	{
	  fprintf(print_fh, "/Courier findfont %g scalefont setfont\n", smallfontsize);
	  for(row2=0; row2<nx; row2++)
	  for(col2=0; col2<ny; col2++)
	  {
	    value=row2*ny+col2+1;
	    if(initialpossible[row][col] & (1<<value))
	    {
	      fprintf(print_fh, "%g %g moveto\n", x+printer_cellsize/ny*(col2+0.5),y-printer_cellsize/nx*(row2+0.5));
	      if(eliminated[row][col] & (1<<value))
	      {
		fprintf(print_fh, "1 0 1 setrgbcolor\n");
	      }
	      else if(possible[row][col] & (1<<value))
	      {
		fprintf(print_fh, "0 0 0 setrgbcolor\n");
	      }
	      else
	      {
		fprintf(print_fh, "0.5 0.5 0.5 setrgbcolor\n");
	      }
	      fprintf(print_fh, "(%c) showcenter\n", symbollist[value]);
	    }
	  }
	}
	x+=printer_cellsize;
      }
      y-=printer_cellsize;
    }
    fprintf(print_fh, "%g setlinewidth\n", printer_thickwidth);
    fprintf(print_fh, "0 0 0 setrgbcolor\n");
    fprintf(print_fh, "newpath\n");
    fprintf(print_fh, "0 0 moveto\n");
    fprintf(print_fh, "0 %g lineto\n", size);
    fprintf(print_fh, "%g %g lineto\n", size, size);
    fprintf(print_fh, "%g 0 lineto\n", size);
    fprintf(print_fh, "closepath\n");
    fprintf(print_fh, "stroke\n");
    npages=1;
    if((progstat==4 || progstat==5) && print_explain)
    {
      /* Print out explanation: */
      unsolve_sudoku();
      loginit();
      solve_sudoku(1);
      logend();
      main_update();
      yspacing = (dev_ymax-size-dev_ymin-2*margin)/(nlines+1);
      if(yspacing>15) yspacing=15;
      yspacing = 15;
      fontsize = yspacing/1.25;
      fprintf(print_fh, "grestore gsave %g %g translate\n", dev_xmin+margin, dev_ymax-margin-yspacing);
      fprintf(print_fh, "/Courier findfont %g scalefont setfont\n", fontsize);
      y=-size;
      for(i=0;i<strlen(loglines);)
      {
        ilen = logtext_en(loglines+i, &buf1);
	i+=ilen;
	fprintf(print_fh, "0 %g moveto (",y);
	for(j=0;j<strlen(buf1);j++)
	{
	  if(buf1[j]=='(' ||
	     buf1[j]==')')
	  {
	    fprintf(print_fh, "\\%c", buf1[j]);
	  }
	  else
	  {
	    fprintf(print_fh, "%c", buf1[j]);
	  }
	}
	fprintf(print_fh, ") show\n");
	y -= yspacing;
	if(y+dev_ymax-yspacing-2*margin<0)
	{
	  fprintf(print_fh, "grestore showpage gsave %g %g translate\n", dev_xmin+margin, dev_ymax-margin-yspacing);
	  fprintf(print_fh, "/Courier findfont %g scalefont setfont\n", fontsize);
	  npages++;
	  y=0;
	}
      }
      fprintf(print_fh, "grestore\n");
    }
    fprintf(print_fh, "showpage\n");
    fprintf(print_fh, "%%%%Pages: %d\n",npages);
    fprintf(print_fh, "%%%%EOF\n%c", (char) 4);

    fclose(print_fh);
  }

  XtFree(filename);
  XtUnrealizeWidget(main_printdialog);
}

static void printer_fileclose_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  XtUnrealizeWidget(main_printdialog);
}

static void printer_papersize_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
  letter = (long) client_data;
}

static void main_wnd_init()
{
  int i;
  Atom atom;
  Widget menubar;
  XGCValues values;
  XtGCMask valueMask;

  XmString s[20];
  XmString stmp;

  n=0;
  XtSetArg(wargs[n], XtNtitle, "Sudoku"); n++;
  XtSetArg(wargs[n], XtNgeometry, "+10+10"); n++;
  main_wnd = XmCreateMainWindow(toplevel, "main", wargs, n);
  atom = XmInternAtom(XtDisplay(toplevel),"WM_DELETE_WINDOW",TRUE);
  XmAddWMProtocolCallback(toplevel, atom, quit_cb, False);
  n=0;
  main_form = XtCreateManagedWidget("main_form", xmFormWidgetClass,
                                    main_wnd, wargs, n);

  /* Menu bar */

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  settings_pb = XtCreateManagedWidget("Settings", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(settings_pb, XmNactivateCallback, settings_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, settings_pb); n++;
  copy_pb = XtCreateManagedWidget("Copy", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(copy_pb, XmNactivateCallback, copy_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, copy_pb); n++;
  paste_pb = XtCreateManagedWidget("Paste", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(paste_pb, XmNactivateCallback, paste_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, paste_pb); n++;
  create_pb = XtCreateManagedWidget("Create", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(create_pb, XmNactivateCallback, create_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, create_pb); n++;
  solve_pb = XtCreateManagedWidget("Show solution", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(solve_pb, XmNactivateCallback, solve_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, solve_pb); n++;
  play_pb = XtCreateManagedWidget("Play", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(play_pb, XmNactivateCallback, play_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, play_pb); n++;
  hint_pb = XtCreateManagedWidget("Hint", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(hint_pb, XmNactivateCallback, hint_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, hint_pb); n++;
  print_pb = XtCreateManagedWidget("Print", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(print_pb, XmNactivateCallback, print_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, print_pb); n++;
  quit_pb = XtCreateManagedWidget("Quit", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(quit_pb, XmNactivateCallback, quit_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, quit_pb); n++;
  about_pb = XtCreateManagedWidget("About", xmPushButtonWidgetClass,
                                      main_form, wargs, n);
  XtAddCallback(about_pb, XmNactivateCallback, about_cb, NULL);

  /* Main drawing area */

  recreate_drawing();
  
  valueMask = GCForeground | GCBackground | GCLineWidth;
  values.foreground = BlackPixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.background = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.line_width = THINWIDTH;
  thin_gc = XtGetGC(drawing_wnd, valueMask, &values);

  valueMask = GCForeground | GCBackground | GCLineWidth;
  values.foreground = BlackPixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.background = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.line_width = THICKWIDTH;
  thick_gc = XtGetGC(drawing_wnd, valueMask, &values);

  font_width=applicationdata.textfont->max_bounds.width;
  font_height=applicationdata.textfont->ascent+applicationdata.textfont->descent;

  smallfont_width=applicationdata.smalltextfont->max_bounds.width;
  smallfont_height=applicationdata.smalltextfont->ascent+applicationdata.smalltextfont->descent;

  valueMask = GCForeground | GCBackground | GCFont;
  values.foreground = applicationdata.blackcolor;
  values.background = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.font=applicationdata.textfont->fid;
  black_gc = XtGetGC(drawing_wnd, valueMask, &values);

  valueMask = GCForeground | GCBackground | GCFont;
  values.foreground = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.background = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.font=applicationdata.textfont->fid;
  background_gc = XtGetGC(drawing_wnd, valueMask, &values);

  valueMask = GCForeground | GCBackground | GCFont;
  values.foreground = applicationdata.redcolor;
  values.background = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.font=applicationdata.textfont->fid;
  red_gc = XtGetGC(drawing_wnd, valueMask, &values);

  valueMask = GCForeground | GCBackground | GCFont;
  values.foreground = applicationdata.blackcolor;
  values.background = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.font=applicationdata.smallboldtextfont->fid;
  small_black_gc = XtGetGC(drawing_wnd, valueMask, &values);

  valueMask = GCForeground | GCBackground | GCFont;
  values.foreground = applicationdata.graycolor;
  values.background = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.font=applicationdata.smalltextfont->fid;
  small_gray_gc = XtGetGC(drawing_wnd, valueMask, &values);

  valueMask = GCForeground | GCBackground | GCFont;
  values.foreground = applicationdata.pinkcolor;
  values.background = WhitePixel(XtDisplay(drawing_wnd),XDefaultScreen(XtDisplay(drawing_wnd)));
  values.font=applicationdata.smalltextfont->fid;
  small_pink_gc = XtGetGC(drawing_wnd, valueMask, &values);

  n=0;
  XtSetArg(wargs[n], XtNtitle, "Sudoku settings"); n++;
  main_settingsdialog = XtCreateApplicationShell("sudoku_settings", topLevelShellWidgetClass, wargs, n);

  n=0;
  settings_form = XtCreateManagedWidget("form", xmFormWidgetClass, main_settingsdialog, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopOffset, 10); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("No. of rows:        "))); n++;
  nx_label = XmCreateLabel(settings_form, "nxTitle", wargs, n);
  XtManageChild(nx_label);
  XmStringFree(s[0]);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopOffset, -8); n++;
  XtSetArg(wargs[n], XmNtopWidget, nx_label); n++;
  XtSetArg(wargs[n], XmNleftWidget, nx_label); n++;
  XtSetArg(wargs[n], XmNorientation, XmHORIZONTAL); n++;
  nx_radio = XmCreateRadioBox(settings_form, "RadioBoxNx", wargs, n);
  XtManageChild(nx_radio);

  for(i=2; i<=MAXNX; i++)
  {
    char t1[10];
    char t2[12];
    sprintf(t1, "%d", i);
    sprintf(t2, "ToggleNx%d", i);
    n=0;
    XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple(t1))); n++;
    XtSetArg(wargs[n], XmNset, (nx==i)?True:False); n++;
    nxradios[i] = XmCreateToggleButton(nx_radio, t2, wargs, n);
    XtManageChild(nxradios[i]);
    XmStringFree(s[0]);
    XtAddCallback(nxradios[i], XmNarmCallback, nx_cb, (XtPointer) i);
  }

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, nx_label); n++;
  XtSetArg(wargs[n], XmNtopOffset, 10); n++;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("No. of columns:     "))); n++;
  ny_label = XmCreateLabel(settings_form, "nyTitle", wargs, n);
  XtManageChild(ny_label);
  XmStringFree(s[0]);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopOffset, -8); n++;
  XtSetArg(wargs[n], XmNtopWidget, ny_label); n++;
  XtSetArg(wargs[n], XmNleftWidget, ny_label); n++;
  XtSetArg(wargs[n], XmNorientation, XmHORIZONTAL); n++;
  ny_radio = XmCreateRadioBox(settings_form, "RadioBoxNy", wargs, n);
  XtManageChild(ny_radio);

  for(i=2; i<=MAXNY; i++)
  {
    char t1[10];
    char t2[12];
    sprintf(t1, "%d", i);
    sprintf(t2, "ToggleNy%d", i);
    n=0;
    XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple(t1))); n++;
    XtSetArg(wargs[n], XmNset, (ny==i)?True:False); n++;
    nyradios[i] = XmCreateToggleButton(ny_radio, t2, wargs, n);
    XtManageChild(nyradios[i]);
    XmStringFree(s[0]);
    XtAddCallback(nyradios[i], XmNarmCallback, ny_cb, (XtPointer) i);
  }

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopOffset, 12); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, ny_label); n++;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("No. of filled cells:"))); n++;
  settings_label = XmCreateLabel(settings_form, "FilledTitle", wargs, n);
  XtManageChild(settings_label);
  XmStringFree(s[0]);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, ny_label); n++;
  XtSetArg(wargs[n], XmNtopOffset, 8); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, settings_label); n++;
  sprintf(filled_buffer, "%d", nfilled);
  XtSetArg(wargs[n], XmNvalue, filled_buffer); n++;
  XtSetArg(wargs[n], XmNmaxLength, sizeof(filled_buffer)-1); n++;
  XtSetArg(wargs[n], XmNcolumns, sizeof(filled_buffer)-1); n++;
  XtSetArg(wargs[n], XmNverifyBell, False); n++;

  settings_edit = XtCreateManagedWidget("settings_edit", xmTextFieldWidgetClass, settings_form, wargs, n);
  XtManageChild(settings_edit);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, settings_edit); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
  separator1 = XtCreateManagedWidget("separator1", xmSeparatorWidgetClass, settings_form, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, separator1); n++;
  settings_radio2 = XmCreateRadioBox(settings_form, "RadioBox2", wargs, n);
  XtManageChild(settings_radio2);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("Not symmetric"))); n++;
  XtSetArg(wargs[n], XmNset, (symmetrical==0)?True:False); n++;
  symmetricals[0] = XmCreateToggleButton(settings_radio2, "ToggleSimple21", wargs, n);
  XtManageChild(symmetricals[0]);
  XmStringFree(s[0]);
  XtAddCallback(symmetricals[0], XmNarmCallback, symmetrical_cb, (XtPointer) 0);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("X&Y mirror"))); n++;
  XtSetArg(wargs[n], XmNset, (symmetrical==1)?True:False); n++;
  symmetricals[1] = XmCreateToggleButton(settings_radio2, "ToggleSimple22", wargs, n);
  XtManageChild(symmetricals[1]);
  XmStringFree(s[0]);
  XtAddCallback(symmetricals[1], XmNarmCallback, symmetrical_cb, (XtPointer) 1);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("180 deg. rotate"))); n++;
  XtSetArg(wargs[n], XmNset, (symmetrical==2)?True:False); n++;
  symmetricals[2] = XmCreateToggleButton(settings_radio2, "ToggleSimple23", wargs, n);
  XtManageChild(symmetricals[2]);
  XmStringFree(s[0]);
  XtAddCallback(symmetricals[2], XmNarmCallback, symmetrical_cb, (XtPointer) 2);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, settings_radio2); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
  separator2 = XtCreateManagedWidget("separator2", xmSeparatorWidgetClass, settings_form, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, separator2); n++;
  settings_radio = XmCreateRadioBox(settings_form, "RadioBox1", wargs, n);
  XtManageChild(settings_radio);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("Very easy"))); n++;
  XtSetArg(wargs[n], XmNset, (level==LEVEL_VERYEASY)?True:False); n++;
  levels[0] = XmCreateToggleButton(settings_radio, "ToggleSimple1", wargs, n);
  XtManageChild(levels[0]);
  XmStringFree(s[0]);
  XtAddCallback(levels[0], XmNarmCallback, level_cb, (XtPointer) LEVEL_VERYEASY);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("Easy"))); n++;
  XtSetArg(wargs[n], XmNset, (level==LEVEL_EASY)?True:False); n++;
  levels[1] = XmCreateToggleButton(settings_radio, "ToggleSimple2", wargs, n);
  XtManageChild(levels[1]);
  XmStringFree(s[0]);
  XtAddCallback(levels[1], XmNarmCallback, level_cb, (XtPointer) LEVEL_EASY);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("Medium (SSC)"))); n++;
  XtSetArg(wargs[n], XmNset, (level==LEVEL_MEDIUM)?True:False); n++;
  levels[2] = XmCreateToggleButton(settings_radio, "ToggleSimple3", wargs, n);
  XtManageChild(levels[2]);
  XmStringFree(s[0]);
  XtAddCallback(levels[2], XmNarmCallback, level_cb, (XtPointer) LEVEL_MEDIUM);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("Hard (SSC+DS)"))); n++;
  XtSetArg(wargs[n], XmNset, (level==LEVEL_HARD)?True:False); n++;
  levels[3] = XmCreateToggleButton(settings_radio, "ToggleSimple4", wargs, n);
  XtManageChild(levels[3]);
  XmStringFree(s[0]);
  XtAddCallback(levels[3], XmNarmCallback, level_cb, (XtPointer) LEVEL_HARD);
  
  n=0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("Difficult (SSC+DS+Brute Force)"))); n++;
  XtSetArg(wargs[n], XmNset, (level==LEVEL_DIFFICULT)?True:False); n++;
  levels[4] = XmCreateToggleButton(settings_radio, "ToggleSimple5", wargs, n);
  XtManageChild(levels[4]);
  XmStringFree(s[0]);
  XtAddCallback(levels[4], XmNarmCallback, level_cb, (XtPointer) 4);
  
  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, settings_radio); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNrightAttachment, XmATTACH_FORM); n++;
  separator3 = XtCreateManagedWidget("separator3", xmSeparatorWidgetClass, settings_form, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, separator3); n++;
  check1 = XmCreateSimpleCheckBox(settings_form, "CheckBox1", wargs, n);
  XtManageChild(check1);

  n = 0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("Eliminate candidates from the beginning"))); n++;
  XtSetArg(wargs[n], XmNset, XmSET); n++;
  settings_smallinitial = XmCreateToggleButton(check1, "ToggleBuffer", wargs, n);
  XtAddCallback(settings_smallinitial, XmNvalueChangedCallback, smallinitial_cb, NULL);
  XtManageChild(settings_smallinitial);
  XmStringFree(s[0]);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, check1); n++;
  check2 = XmCreateSimpleCheckBox(settings_form, "CheckBox2", wargs, n);
  XtManageChild(check2);

  n = 0;
  XtSetArg(wargs[n], XmNlabelString, (s[0]=XmStringCreateSimple("Eliminate candidates automatically"))); n++;
  XtSetArg(wargs[n], XmNset, XmSET); n++;
  settings_autosmall = XmCreateToggleButton(check2, "ToggleBuffer", wargs, n);
  XtAddCallback(settings_autosmall, XmNvalueChangedCallback, autosmall_cb, NULL);
  XtManageChild(settings_autosmall);
  XmStringFree(s[0]);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNbottomAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, check2); n++;
  XtSetArg(wargs[n], XmNtopOffset, 10); n++;
  XtSetArg(wargs[n], XmNbottomOffset, 10); n++;
  settings_ok = XtCreateManagedWidget("OK", xmPushButtonWidgetClass, settings_form, wargs, n);
  XtAddCallback(settings_ok, XmNactivateCallback, settingsok_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, settings_ok); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, settings_ok); n++;
  settings_cancel = XtCreateManagedWidget("Cancel", xmPushButtonWidgetClass, settings_form, wargs, n);
  XtAddCallback(settings_cancel, XmNactivateCallback, settingscancel_cb, NULL);

  n=0;
  XtSetArg(wargs[n], XtNtitle, "PostScript File Selection"); n++;
  main_printdialog = XtCreateApplicationShell("main_printdialog", topLevelShellWidgetClass, wargs, n);

  n=0;
  printer_fileform = XtCreateManagedWidget("form", xmFormWidgetClass, main_printdialog, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNpattern, (stmp=XmStringCreateSimple("*.ps"))); n++;
  printer_fileselection = XtCreateManagedWidget("printer_fileselection", xmFileSelectionBoxWidgetClass, printer_fileform, wargs, n);
  XmStringFree(stmp);

  XtAddCallback(printer_fileselection, XmNokCallback, printer_fileopen_cb, (XtPointer)NULL);
  XtAddCallback(printer_fileselection, XmNcancelCallback, printer_fileclose_cb, (XtPointer)NULL);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(wargs[n], XmNtopOffset, 5); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  papersize_label = XtCreateManagedWidget("Select paper size", xmLabelWidgetClass, printer_fileform, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, papersize_label); n++;
  XtSetArg(wargs[n], XmNleftOffset, 0); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  papersize = XmCreateOptionMenu(printer_fileform, "papersize", wargs, n);

  n=0;
  papersize_pane = XmCreatePulldownMenu(papersize, "pulldown", wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateLocalized("A4"))); n++;
  A4_w = XtCreateManagedWidget("A4_select", xmPushButtonWidgetClass, papersize_pane, wargs, n);
  XtAddCallback(A4_w, XmNactivateCallback, printer_papersize_cb, (XtPointer) 0);

  n=0;
  XtSetArg(wargs[n], XmNlabelString, (stmp=XmStringCreateLocalized("Letter"))); n++;
  Letter_w = XtCreateManagedWidget("letter_select", xmPushButtonWidgetClass, papersize_pane, wargs, n);
  XtAddCallback(Letter_w, XmNactivateCallback, printer_papersize_cb, (XtPointer) 1);

  n=0;
  XtSetArg(wargs[n], XmNsubMenuId, papersize_pane); n++;
  XtSetValues(papersize, wargs, n);

  XtManageChild(papersize);

  n=0;
  XtSetArg(wargs[n], XmNmenuHistory, letter?Letter_w:A4_w); n++;
  XtSetValues(papersize, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, papersize); n++;
  XtSetArg(wargs[n], XmNtopOffset, 5); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  printer_explain_label = XtCreateManagedWidget("Print explanation", xmLabelWidgetClass, printer_fileform, wargs, n);

  n=0;
  XtSetArg(wargs[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNtopWidget, printer_explain_label); n++;
  XtSetArg(wargs[n], XmNleftOffset, 10); n++;
  XtSetArg(wargs[n], XmNleftAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(wargs[n], XmNleftWidget, printer_fileselection); n++;
  printer_explain_check = XmCreateSimpleCheckBox(printer_fileform, "printer_explain_check", wargs, n);
  XtManageChild(printer_explain_check);

  n=0;
#if XmVERSION==1
  XtSetArg(wargs[n], XmNindicatorOn, True); n++;
  XtSetArg(wargs[n], XmNset, print_explain?True:False); n++;
#else
  XtSetArg(wargs[n], XmNindicatorOn, XmINDICATOR_CHECK_BOX); n++;
  XtSetArg(wargs[n], XmNset, print_explain?XmSET:XmUNSET); n++;
#endif
  XtSetArg(wargs[n], XmNindicatorSize, 20); n++;
  printer_explain_select = XmCreateToggleButton(printer_explain_check, "", wargs, n);
  XtManageChild(printer_explain_select);
  XmFileSelectionDoSearch(printer_fileselection, NULL);
  XtManageChild(main_wnd);
}

int test_stop()
{
  int status;
  XEvent event;

  while(status=XtPending())
  {
    XtNextEvent(&event);
    XtDispatchEvent(&event);
  }
  return stop_pressed;
}

static void sudokumain()
{
  XEvent event;
  int status;

AGAIN:
  status = XtPending();
  if(status)
  {
AGAIN2:
    XtNextEvent(&event);
    XtDispatchEvent(&event);
    goto AGAIN;
  }
  if(progstat==1 && !stop_pressed)
  {
    sudoku_run();
    goto AGAIN;
  }
  goto AGAIN2;
}  

main(int argc, char **argv)
{
  char *env;
  unsigned int seed;

  env=getenv("SUDOKU_SEED");
  if(env)
  {
    seed=atoi(env);
  }
  else
  {
    seed=time(NULL);
  }

  srandom(seed);

  toplevel=XtInitialize("Sudoku", "Sudoku", NULL, 0, &argc, argv);
  XtVaSetValues(toplevel, XtNallowShellResize, True, NULL);

  XtGetApplicationResources(toplevel, &applicationdata,
                            resources, XtNumber(resources), NULL, 0);

  init_bitmaps(nxny);
  main_wnd_init();

  progstat = 0;

  XtRealizeWidget(toplevel);

  update_menu();

  sudokumain();
}
